# ICFP 2023 Artifact

Name: *Verifying Reliable Network Components in a Distributed Separation Logic
       with Dependent Separation Protocols*

## Artifact Instructions

### Organization

This source code package contains the following elements:
  - *README.md*: this file.

  - *ocaml2lang*: folder containing the compiler from OCaml code into Aneris-lang.

  - *aneris*: folder containing the Aneris project, a Distributed Separation
    Logic built on top of the Iris framework. It contains the definition of
    Aneris-Lang (a small ML-like language featuring sockets and network
    communication primitives), the instantiation of the Iris logic for Aneris,
    as well as the proof of adequacy. The project also contains some examples of
    verified using the Aneris logic (many of such case studies are actually
    Aneris-Lang programs generated using our OCaml to Aneris-Lang compiler).

### Installation

#### Aneris

1. In the `aneris` folder, run
     ``make``
   to compile the Aneris project.

The Aneris project `README.md` contains complementary installation instructions
in the case one wishes to clone the project directly from GitHub. In particular,
it contains instructions on how to recursively obtain some needed dependencies
for Aneris.

#### Compiler from OCaml into Aneris-lang

1. Create a switch for the `ocaml-variants.4.12.0+domains` variant, as follows:
     ``opam switch create ocaml-variants.4.12.0+domains``

2. In the `ocaml2lang` folder, pin the `ocaml2lang` package:
     ``opam pin add .``

3. This should install the `ocaml2lang` package in your machine and produce
   the `o2a` executable. You can test it by doing:
     ``o2a --help``

### Overview of the OCaml to Aneris-Lang Architecture

The `ocaml2lang` folder is organized in following sub-folders:
```
    ocaml2lang
    |
    |__bin
    |
    |__lib
```
Other than the mentioned sub-folders, `ocaml2lang` contains a set of files that
are normally found in an OCaml project that runs on top of `opam` and `dune`.

  * `bin`:
    This folder contains a single OCaml file, `main.ml`, used as the entry point
    of the compiler tool. It basically sets some command-line options, opens an
    reading channel for the input file, and finally calls pretty-printing
    functions for the generated Coq code.

  * `lib`:
    This folder contains the core of the compiler implementation. Let us briefly
    describe the source files contained in this folder:
      - `ast.ml/ast.mli`: definition of our intermediate representation of
        Aneris-Lang implementations as Coq terms.
      - `ml_project.ml`: encoding of the information from the `_OCamlProject`.
      - `mllexer.mly\mlparser.mly`: lexer and parser (respectively) for reading
        the contents of an `_OCamlProject` file.
      - `pp_aneris.ml`: several pretty-printing functions to print Coq files
        containing Aneris-Lang implementations.
      - `translate_aneris.ml`: the core functionality of the compiler, as it
        defines the translation functions from OCaml implementations into
        Aneris-Lang terms.

### Step-by-step instructions on exercising the artifact

#### Compiling from OCaml sources to Aneris-lang

The compiler from OCaml into Aneris-lang (the `ocaml2lang` package) is already
installed in the virtual machine. One can check it bu running
    ``o2a --help``

This prints a common usage message for the tool.

In order to properly interact with the compiler, i.e., generate Aneris-lang
implementations from OCaml code, do the following:

1. Navigate to the root folder of the `aneris` project.

2. Run `o2a --rewrite`. This will generate Coq files for the various OCaml
   programs contained in the folder `ml_sources`. It produces the following output:
   ```
   Created: _generated
   Created: aneris/examples/reliable_communication/lib/repdb/repdb_code.v
   Created: aneris/aneris_lang/lib/nodup_code.v
   Created: aneris/aneris_lang/lib/queue_code.v
   Created: aneris/aneris_lang/lib/coin_flip_code.v
   Created: aneris/aneris_lang/lib/map_code.v
   Created: aneris/aneris_lang/lib/set_code.v
   Created: aneris/aneris_lang/lib/vector_clock/vector_clock_code.v
   Created: aneris/aneris_lang/lib/serialization/serialization_code.v
   Created: aneris/aneris_lang/lib/network_util_code.v
   Created: aneris/examples/reliable_communication/client_server_printing.v
   Created: aneris/examples/reliable_communication/examples/repdb_leader_followers/causality_example_code.v
   Created: aneris/aneris_lang/lib/par_code.v
   Created: aneris/examples/reliable_communication/examples/messages_in_order_loop/messages_in_order_loop_code.v
   Created: aneris/aneris_lang/lib/spawn_code.v
   Created: aneris/examples/reliable_communication/lib/repdb/log_code.v
   Created: aneris/aneris_lang/lib/list_code.v
   Created: aneris/examples/reliable_communication/lib/mt_server/mt_server_code.v
   Created: aneris/aneris_lang/lib/bag_code.v
   Created: aneris/examples/reliable_communication/lib/dlm/dlm_code.v
   Created: aneris/examples/reliable_communication/client_server_code.v
   ```

This stands for the list of generated Coq files. The file `_generated` (printed
on the same directory where `o2a --rewirte` is called) contains the exact same
list of files.

#### The _OCamlProject Configuration File

For the `o2a` tool to work properly, one must provide some input/output
instructions. These are given in the `_OCamlProject` file, which is roughly
inspired by how `_CoqProject` works within Coq projects.

An `_OCamlProject` file is divided in four distinct sections:

  * `OUTPUT_ROOT`:
    Each line in this section establishes a mapping from the name of the OCaml
    sources root folder into the root folder of the generated Coq files.

  * `IMPORT_PATH_PREFIX`:
    Each line in this section establishes a mapping from the name of an OCaml
    directory into a name that shall be used in `From ... Require Import ...`
    lines in the generated Coq file.

  * `ML_SOURCES`:
    The list of OCaml files to be translated.

  * `ML_DEPENDENCIES`:
    A list of dependencies.

For a better understanding, let us examine the `_OCamlProject` that we have
ourselves to build this artifact. This contains the following:
```
OUTPUT_ROOT:
	ml_sources aneris

IMPORT_PATH_PREFIX:

	ml_sources aneris

ML_SOURCES:

	aneris_lang/lib/network_util_code.ml
	aneris_lang/lib/list_code.ml
    aneris_lang/lib/queue_code.ml
	aneris_lang/lib/set_code.ml
	aneris_lang/lib/map_code.ml
	aneris_lang/lib/bag_code.ml
	aneris_lang/lib/nodup_code.ml
	aneris_lang/lib/coin_flip_code.ml
	aneris_lang/lib/serialization/serialization_code.ml
	aneris_lang/lib/vector_clock/vector_clock_code.ml
	aneris_lang/lib/spawn_code.ml
	aneris_lang/lib/par_code.ml
	examples/reliable_communication/client_server_printing.ml
	examples/reliable_communication/client_server_code.ml
  	examples/reliable_communication/lib/dlm/dlm_code.ml
	examples/reliable_communication/lib/mt_server/mt_server_code.ml
	examples/reliable_communication/lib/repdb/log_code.ml
	examples/reliable_communication/lib/repdb/repdb_code.ml
  	examples/reliable_communication/examples/dlm_db_example/dlm_db_example_code.ml
	examples/reliable_communication/examples/messages_in_order_loop/messages_in_order_loop_code.ml

ML_DEPENDENCIES:

    aneris_lang
	aneris_lang/lib
    aneris_lang/lib/serialization
	aneris_lang/lib/vector_clock
	examples/reliable_communication/
	examples/reliable_communication/lib/
	examples/reliable_communication/lib/dlm/
	examples/reliable_communication/lib/mt_server/
	examples/reliable_communication/lib/repdb/
```

We explain now each section of this file:

  * `OUTPUT_ROOT`:
    The root folder of the OCaml sources is named `ml_sources`. The `o2a` tool crates
    a new folder named `aneris` and it will populate this folder with the exact
    same path structure as found in the `ml_sources` folder. For instance, the
    `client_server_code.ml` (we give more details about the translation of this
    file later) is located under
    `aneris/ml_sources/examples/reliable_communication/`. Since we run `o2a
    --rewrite` in the `aneris` folder, the following path structure is created:
```
    aneris/aneris/examples/reliable_communication/
```

  * `IMPORT_PATH_PREFIX`:
    Given the newly created path inside the `aneris` folder, the tool must
    correctly take into account such information when setting the importing
    information in the generated Coq files. For the case of the
    `client_server_code.v` file, it starts with:
```
    From aneris.aneris_lang Require Import ast.
    From aneris.aneris_lang.lib Require Import queue_code.
    From aneris.aneris_lang.lib Require Import map_code.
    From aneris.aneris_lang.lib.serialization Require Import serialization_code.
    From aneris.aneris_lang.lib Require Import network_util_code.
    From aneris.examples.reliable_communication Require Import client_server_printing.
```

  * `ML_SOURCES`:
    The list of all case studies.

  * `ML_DEPENDENCIES`:
    The list of all necessary dependencies.

#### Example: client_server_code

Let us pick the example of the `client_server_code.ml` OCaml implementation file
to check the relation between the original code and the generated one. This file
contains several OCaml functions that describe the connection protocol between a
client and server. For instance, the `connect`, `send`, `try_recv`, and `recv`
functions form the backbone of this implementation. Once we run the `o2a` tool,
the Coq file `client_server_code.v` is generated (under
`aneris/examples/reliable_communication/`). Basically, this file contains the
Aneris-Lang translation of every previous OCaml function (comments are also
kept). For instance, in the original `.ml` file, the `send` function is
implemented (in OCaml) as follows:

```
let send (c : ('a, 'b) chan_descr) (mbody : 'a) =
  let sdata = fst c in
  let (sbuf, smon) = sdata in
  monitor_acquire smon;
  let qe = !sbuf in
  sbuf := queue_add mbody qe;
  monitor_signal smon;
  monitor_release smon
```
while in the `client_server_code.v` file it is implemented (in Aneris-Lang) as
follows:
```
Definition send : val :=
  λ: "c" "mbody",
  let: "sdata" := Fst "c" in
  let: "sbuf" := Fst "sdata" in
  let: "smon" := Snd "sdata" in
  monitor_acquire "smon";;
  let: "qe" := ! "sbuf" in
  "sbuf" <- (queue_add "mbody" "qe");;
  monitor_signal "smon";;
  monitor_release "smon".
```
One can easily check the two implementations share many common traits, e.g., the
use of the same identifiers (in Aneris-Lang variable names are expressed within
quotes) and similar code structure (`let_in` expressions are written in
Aneris-Lang as `let: _ := _ in`; sequence of instructions is represented using
the `;;` operator).

For every other example contained in this artifact, the generated Aneris-Lang
implementation closely resembles that of the OCaml code.

Finally, to derive a formal proof of the generated Aneris-Lang implementation,
it is up to the user to write her own, new Coq file. For the case of the
`send` function from the `client_server_code` file, the proof can be found under
`aneris/aneris/examples/reliable_communication/proof/common_user/proof_of_send.`

### Differences between the mechanization and the paper

There are a number of small differences between the paper presentation of RClib
and its examples and the formalization in Coq, that are briefly discussed here.

**Names**

Various names of definitions/quantifiers do not correspond exactly between
the paper and the mechanization.
The most notable difference is the RPC framework, which is called mt_server in
the mechanization.
We believe that the correspondence between the remaining names can be
interpreted from context.

**Invariant Namespaces**

The paper specifications of the framework and the session escrow pattern
omit Iris invariant namespaces, while the mechanization includes a namespace
parameter as a part of the user parameters `RCParams_srv_N : namespace`,
and in the session escrow construction.

Namespaces enforce that Iris invariants can only be opened once per
atomic step, as opening the same invariant twice results in an unsound logic.
With our current framework specifications, we would have no use for opening
multiple invariants per step, as the specifications do not allow for atomic
ghost transitions, and thus it would be sufficient to pick a static namespace
for all invariants (in other words, omit the parameter).
The parameter of the mechanization is in place to allow for potential future
extensions of the framework which might allow have atomic ghost transitions.

**Step modality vs pre-step modality**

In the paper we present the step modality, used in the session escrow pattern.
In the mechanisation, we use a so-called "pre-step" modality, which can be
resolve _before_ taking a step.
In particular, this is used in the sescrow-init rule of the session escrow pattern.
The pre-step modality was omitted in the paper for brevity sake.
Our proofs could trivially be obtained with the step modality, as one could simply
allocate a new session escrow pattern one step earlier.

**Injectivity of serializers**

The mechanization includes proofs of injectivity of the chosen serializers as
a part of the user parameters. We omitted this detail in the paper for brevity's
sake.

**Lock invariant definitions**

The lock invariant definitions presented in the paper differ syntactically from
those in the mechanization, although the semantic meaning is preserved.

*Queue resources*

In the paper we use a "queue resource" `l ↦q{ip} b`, which semantically
correspond to `∃ v. l ↦{ip} v ∗ is_queue v b`.
In the mechanization, we inline this definition, which results in
one more existential quantifier `q` in the definition of both the send
and receive lock invariants.

*Ghost and reference arithmetic*

We use ghost state to relate the receive counter `m` of the endpoint ownership,
to the acknowledgement id `aid` of the receive lock invariant.
The quantifications and arithmetics involved in this relation differs
slightly between the paper and the mechanization.

In the paper, we quantify over the acknowledgement id `aid`
and assert that its respective reference governs it `l_aid ↦{ip} aid`.
We orthogonally track that the receive index is offset by the
acknowledgement id, relative to the buffer size `∘(aid - |rbuf|)`.

In the mechanization we instead quantify over the receive index
`ridx`, and assert that it agrees with the protocol index via ghost state `∘ ridx`.
We then compute the acknowledgement id as an offset of the receive idx
`l_aid ↦{ip} (ridx + length vs)`.

The two correspond semantically, but we chose the former for the paper for
intuitive purposes, while the latter was easier to reason about in Coq.

*Dependent specification pattern in the paper and typeclasses in Coq*

In the paper, we use so-called dependent specification pattern to present the
specs for reliable communication library and the RPC library. In this pattern,
the specifications of the API primitives are dependent on the user parameters,
eg., `(UP : RC_UserParams)`, provided by the user, and on the abstract
specification parameters eg., `(𝑆 : RC_Resources UP)` provided by the library
itself. In the Coq mechanization, this is realized using typeclasses for
the API specs declaration and typeclass instances for concrete use of the specs.
For instance, reliable communication dependent specification pattern is
embodied in the mechanization by the following typeclass (where
`Reliable_communication_service_params` corresponds to `RC_Params`):

```
Class Reliable_communication_init := {
      Reliable_communication_init_setup
        E (UP : Reliable_communication_service_params):
      ↑RCParams_srv_N ⊆ E →
      ⊢ |={E}=>
        ∃ ( _ : Chan_mapsto_resource),
        ∃ (SnRes : SessionResources UP),
          SrvInit ∗
          ⌜make_client_skt_spec UP SnRes⌝ ∗
          ⌜make_server_skt_spec UP SnRes⌝ ∗
          ⌜connect_spec UP SnRes⌝ ∗
          ⌜server_listen_spec UP SnRes⌝ ∗
          ⌜accept_spec UP SnRes⌝ ∗
          ⌜send_spec⌝ ∗
          ⌜send_spec_tele⌝ ∗
          ⌜try_recv_spec⌝ ∗
          ⌜recv_spec⌝
    }.
```

Note that in this particular case, in addition to `SessionResources UP`, the library also provides
existentially quantified `( _ : Chan_mapsto_resource)` resource for the channel
endpoint points-to prediate. This is merely a current design (and could be done outside
the class) - the reason for this is that a concrete way we used ghost names to
track in the logic client identities known by the server resulted in using some of those names
in the channel endpoint points-to prediate as well, but that can be improved and
is not of fundamental significance.

### Claims:

The paper includes a feature-realistic implementation of reliable communication
The paper verifies the safety of the implementation, and provides high-level Actris-style specifications
The specifications support vertical modularity
The verification included non-trivial proofs, which we resolved with the novel "session escrow pattern"
To substantiate these claims we refer to:

a. The implementation of RClib

	`/aneris/aneris/ml_sources/examples/reliable_communication/client_server_code.ml`

	`/aneris/aneris/ml_sources/examples/reliable_communication/client_server_code.mli`

b. The mechanisation of RClib

	Specs:
	`/aneris/aneris/examples/reliable_communication/spec/`

	Resource definitions and properties:
	`/aneris/aneris/examples/reliable_communication/resources/`

	Proofs of the specs:
	`/aneris/aneris/examples/reliable_communication/proof/`

	Closed Proof of instantiation of the typeclasses (dependent specification pattern):
	`/aneris/aneris/examples/reliable_communication/proof/`

c. The mechanisation of RPC on top of RCLib

	Ocaml source files:

	`/aneris/ml_sources/examples/reliable_communication/lib/mt_server/mt_server_code.ml`

	`/aneris/ml_sources/examples/reliable_communication/lib/mt_server/mt_server_code.mli`

	Specs:
	`/aneris/aneris/examples/reliable_communication/lib/mt_server/user_params.v`
	`aneris/aneris/examples/reliable_communication/lib/mt_server/spec/api_spec.v#L59`

	Proof:
	`/aneris/aneris/examples/reliable_communication/lib/mt_server/proof/mt_server_proof.v`

d. The mechanisation of the session escrow pattern:

    Step modality:
	`/aneris/aneris/aneris_lang/program_logic/step_update.v#L102`

    Session Escrow Pattern:
	`/aneris/aneris/examples/reliable_communication/resources/session_escrow.v`

    Proof of send and receive specs using session escrow pattern:
	`/aneris/aneris/examples/reliable_communication/proof/common_user/proof_of_recv.v`

	`/aneris/aneris/examples/reliable_communication/proof/common_user/proof_of_send.v`


### Correspondance between the definitions and claims in the paper and the mechanization

**Section 1 INTRODUCTION**

* (L50) Example (OCaml source file): server returning the length of incoming strings:

	`/aneris/ml_sources/examples/reliable_communication/examples/messages_in_order_loop/messages_in_order_loop_code.ml`

* (L50) Example (Coq file generated by the compiler): server returning the length of incoming strings:

	`/aneris/aneris/examples/reliable_communication/examples/messages_in_order_loop/messages_in_order_loop_code.v`

* (L138) echo prot definition

	`/aneris/aneris/examples/reliable_communication/examples/messages_in_order_loop/messages_in_order_loop_proof.v#L21`

* (L146) channel descriptor resource

	`/aneris/aneris/examples/reliable_communication/resources/chan_endpoints_resources.v#L44`

   `/aneris/aneris/examples/reliable_communication/resources/chan_endpoints_resources.v#L107` (notation)

**Section 2 PRIOR WORK AND ITS LIMITATIONS**

* (L271) Fig. 2. The grammar and a selection of rules of the Aneris communication layer.

	`/aneris/aneris/aneris_lang/program_logic/aneris_lifting.v`

* (L304) Theorem 2.1 (Adeqacy of Aneris)

	`/aneris/aneris/aneris_lang/program_logic/aneris_lang/adequacy.v#L222`

**Section 3 RELIABLE COMMUNICATION LIBRARY API AND SPECIFICATION**

* (L502) Fig. 4. The API of the reliable communication library.

	`/aneris/aneris/ml_sources/examples/reliable_communication/client_server_code.mli`

* (L578) Fig. 5. The specifications of the Reliable Communication Library

	`/aneris/aneris/examples/reliable_communication/spec/api_spec.v`

* (600) 3.3 A Simple Example: Verifying a String Length Server

	`/aneris/ml_sources/examples/reliable_communication/examples/messages_in_order_loop/messages_in_order_loop_code.ml`

**Section 4 IMPLEMENTING AND VERIFYING THE RELIABLE COMMUNICATION LIBRARY**

-- **VERIFICATION**

* (L751) The session escrow pattern and step modality.

	`/aneris/aneris/aneris_lang/program_logic/step_update.v#L102` (modality)

	`/aneris/aneris/examples/reliable_communication/resources/session_escrow.v` (SEP)

* (L795) Verifying the connection step

	a) Client side:

	`aneris/examples/reliable_communication/proof/client/proof_of_connect_step_1.v`

	`aneris/examples/reliable_communication/proof/client/proof_of_connect_step_2.v`

	`aneris/examples/reliable_communication/proof/client/proof_of_connect.v`

	b) Server side

	`aneris/examples/reliable_communication/proof/server/proof_of_server_conn_step_to_open_new_conn.v`

	`aneris/examples/reliable_communication/proof/server/proof_of_server_conn_step_to_establish_conn.v`

	`aneris/examples/reliable_communication/proof/server/proof_of_server_listen.v`

* (L804) Verifying the channel descriptors

	`/aneris/aneris/examples/reliable_communication/proof/common_user/proof_of_recv.v`

	`/aneris/aneris/examples/reliable_communication/proof/common_user/proof_of_send.v`

	`/aneris/aneris/examples/reliable_communication/proof/common_protocol/proof_of_make_new_channel_descr.v`


* (L811) `Sbuf_inv` and `rbuf_inv` definitions

	`/aneris/aneris/examples/reliable_communication/resources/chan_endpoints_resources.v#L19` (sbuf)

	`/aneris/aneris/examples/reliable_communication/resources/chan_endpoints_resources.v#L32` (rbuf)


* (L821) `chan_descr` definition

	`/aneris/aneris/examples/reliable_communication/resources/chan_endpoints_resources.v#L44`

* (L839) Verifying the internal network procedures.

	`/aneris/aneris/examples/reliable_communication/proof/common_protocol/proof_of_recv_on_chan.v`

	`/aneris/aneris/examples/reliable_communication/proof/common_protocol/proof_of_send_from_chan_loop.v`

* (L845) Socket protocols

	`/aneris/aneris/examples/reliable_communication/resources/socket_interp.v#L58`

	`/aneris/aneris/examples/reliable_communication/resources/socket_interp.v#L108`

**Section 5 REMOTE PROCEDURE CALL LIBRARY**

* (L884) API

	`/aneris/ml_sources/examples/reliable_communication/lib/mt_server/mt_server_code.mli`

* (L913) Specifications for the RPC library

	`/aneris/aneris/examples/reliable_communication/lib/mt_server/user_params.v` (RPC User Parameters)

	`aneris/aneris/examples/reliable_communication/lib/mt_server/spec/api_spec.v#L59` (RPC Specs)

* (L947) The reliable communication of the RPC library proof sketch

	`/aneris/aneris/examples/reliable_communication/lib/mt_server/proof/mt_server_proof.v`

**Section 6 LAZY REPLICATION WITH LEADER-FOLLOWERS**


* (L1051) The tables giving a summary of the building blocks used in the specification of leader and followers:


	`/aneris/aneris/examples/reliable_communication/lib/repdb/spec/time.v`

	`/aneris/aneris/examples/reliable_communication/lib/repdb/events.v`

	`/aneris/aneris/examples/reliable_communication/lib/repdb/resources`


* (L1089) Fig. 9. The specification for the write and the read operation for both the leader and followers

	`/aneris/aneris/examples/reliable_communication/lib/repdb/spec/api_spec.v`

* (L1099) Fig. 10. Example Client of Leader-Followers

	`/aneris/aneris/examples/reliable_communication/examples/repdb_leader_followers`

* (L1128) Verification of the Leader-Followers KVS

	`/aneris/aneris/examples/reliable_communication/lib/repdb/proof`

### Additional software installed in the virtual machine

Other than the projects required to check the claims of the paper, we have
installed the following additional software:
  - `emacs`
  - `opam`

## QEMU Instructions

The ICFP 2023 Artifact Evaluation Process is using a Debian QEMU image as a base
for artifacts. The Artifact Evaluation Committee (AEC) will verify that this
image works on their own machines before distributing it to authors.  Authors
are encouraged to extend the provided image instead of creating their own. If it
is not practical for authors to use the provided image then please contact the
AEC co-chairs before submission.

QEMU is a hosted virtual machine monitor that can emulate a host processor
via dynamic binary translation. On common host platforms QEMU can also use
a host provided virtualization layer, which is faster than dynamic binary
translation.

QEMU homepage: https://www.qemu.org/

### Installation

#### OSX
``brew install qemu``

#### Debian and Ubuntu Linux
``apt-get install qemu-kvm``

On x86 laptops and server machines you may need to enable the
"Intel Virtualization Technology" setting in your BIOS, as some manufacturers
leave this disabled by default. See Debugging.md for details.


#### Arch Linux

``pacman -Sy qemu``

See the [Arch wiki](https://wiki.archlinux.org/title/QEMU) for more info.

See Debugging.md if you have problems logging into the artifact via SSH.


#### Windows 10

Download and install QEMU via the links at

https://www.qemu.org/download/#windows.

Ensure that `qemu-system-x86_64.exe` is in your path.

Start Bar -> Search -> "Windows Features"
          -> enable "Hyper-V" and "Windows Hypervisor Platform".

Restart your computer.

#### Windows 8

See Debugging.md for Windows 8 install instructions.

### Startup

The base artifact provides a `start.sh` script to start the VM on unix-like
systems and `start.bat` for Windows. Running this script will open a graphical
console on the host machine, and create a virtualized network interface.
On Linux you may need to run with `sudo` to start the VM. If the VM does not
start then check `Debugging.md`

Once the VM has started you can login to the guest system from the host.
Whenever you are asked for a password, the answer is `password`. The default
username is `artifact`.

```
$ ssh -p 5555 artifact@localhost
```

You can also copy files to and from the host using scp.

```
$ scp -P 5555 artifact@localhost:somefile .
```

### Shutdown

To shutdown the guest system cleanly, login to it via ssh and use

```
$ sudo shutdown now
```

### Artifact Preparation

Authors should install software dependencies into the VM image as needed,
preferably via the standard Debian package manager. For example, to install
GHC and cabal-install, login to the host and type:

```
$ sudo apt update
$ sudo apt install ghc
$ sudo apt install cabal-install
```

If you really need a GUI then you can install X as follows, but we prefer
console-only artifacts whenever possible.

```
$ sudo apt install xorg
$ sudo apt install xfce4   # or some other window manager
$ startx
```

See Debugging.md for advice on resolving other potential problems.

If your artifact needs lots of memory you may need to increase the value
of the `QEMU_MEM_MB` variable in the `start.sh` script.

When preparing your artifact, please also follow the [Submission
Guidelines](https://icfp23.sigplan.org/track/icfp-2023-artifact-evaluation#Submission-Guidelines).
