From iris.algebra Require Import excl.
From iris.algebra.lib Require Import excl_auth.
From trillium.prelude Require Import finitary.
From iris.base_logic.lib Require Import invariants.
From iris.proofmode Require Import coq_tactics reduction spec_patterns tactics.
From aneris.aneris_lang Require Import lang.
From aneris.aneris_lang.lib Require Import inject lock_proof list_proof pers_socket_proto.
From aneris.aneris_lang.lib.serialization Require Import serialization_proof.
From aneris.aneris_lang.program_logic Require Import aneris_lifting.
From aneris.examples.reliable_communication Require Import user_params client_server_code.
From aneris.examples.reliable_communication.spec Require Import resources api_spec.
From aneris.examples.reliable_communication.resources Require Import prelude.
From aneris.examples.reliable_communication.instantiation
     Require Import instantiation_of_resources.

Set Default Proof Using "Type".

(* -------------------------------------------------------------------------- *)
(** Instantiation of the server side specifications socket and connect. *)
(* -------------------------------------------------------------------------- *)

From aneris.examples.reliable_communication.proof.common_user Require Import
     proof_of_send
     proof_of_recv.

Section Send_Recv_API_spec_instantiation.
  Context `{!anerisG Mdl Σ}.
  Context `{!lockG Σ}.
  Context `{!chanG Σ}.
  Context `{!server_ghost_names}.
  Context `{User_params: !Reliable_communication_service_params}.

  Implicit Types p : iProto Σ.
  Implicit Types TT : tele.

  Lemma send_spec_holds :
    send_spec.
  Proof.
    rewrite /send_spec.
    iIntros (c v p ip serA Φ) "(H1 & H2) HΦ".
    iApply (send_spec_internal _ _ p with "[$][HΦ]").
    iNext. iIntros "Hc". iApply "HΦ". iFrame.
  Qed.

  Lemma send_tele_spec_holds :
    send_spec_tele.
  Proof.
    rewrite /send_spec_tele.
    iIntros (TT c tt v P p ip serA Φ) "(H1 & Hp & H2) HΦ".
    iApply (send_spec_tele_internal _ _ _ _ _ _ with "[-HΦ][HΦ]").
    - by iFrame.
    - iNext. iIntros "Hc". iApply "HΦ". iFrame.
  Qed.

  Lemma try_recv_spec_holds :
    try_recv_spec.
  Proof.
    rewrite /try_recv_spec.
    iIntros (TT c v P p ip ser Φ) "H1 HΦ".
    iApply (try_recv_spec_internal _ _ _  with "[$][HΦ]").
    iNext. iIntros (w) "Hc". iApply "HΦ".
    iDestruct "Hc" as "[(-> & Hc)|(%tt & -> & Hc & Hp)]".
    - iLeft. iSplit; first done. iFrame.
    - iRight.
      iExists _.
      iSplit; first done.
      iFrame.
  Qed.

  Lemma recv_spec_holds :
    recv_spec.
  Proof.
    rewrite /recv_spec.
    iIntros (TT c v P p ip ser Φ) "H1 HΦ".
    iApply (recv_spec_internal _ _ _  with "[$][HΦ]").
    iNext. iIntros (w) "(Hc & Hp)". iApply "HΦ". iFrame.
  Qed.

End Send_Recv_API_spec_instantiation.
